proc cp { args } {
    # If there's no matching files in a glob, don't try to copy.  (Need to
    # catch glob results to do so)
    set source [lrange $args 0 [expr [llength $args]-2]]
    set dest [lindex $args end]
    file mkdir $dest
    foreach entry $source {
	if { ![catch {glob $entry}] } {
	    foreach file [glob $entry] {
		if { [file isdirectory $file] } {
		    cp $file/* $dest/[file tail $file]
		} else {
		    file copy -force $file $dest
		}
		exec chmod -f u+r $dest/[file tail $file]
	    }
	}
    }
}

proc log { args } {
    .top.log configure -state normal
    foreach arg $args {
	.top.log insert end $arg\n
    }
    .top.log configure -state disabled
    .top.log see end
}

# Set up the installation window.

wm title . "QuIRC Installation System - Log"
frame .top
pack [text .top.log -state disabled -yscrollcommand ".top.scroll set"] -side left
pack [scrollbar .top.scroll -command ".top.log yview"] -fill y -expand 1
pack .top
pack [button .ok -command {destroy .} -text "Ok"]
bind . <Return> {destroy .}
bind . <Down> {.top.log yview scroll 1 units}
bind . <Up> {.top.log yview scroll -1 units}

# Set up the relatively pathed file list from the data dir.

set filelist ""
foreach file [glob $DATADIR/*.tcl] {
    lappend filelist [file tail $file]
}
foreach file [glob $DATADIR/themes/*.tcl] {
    lappend filelist themes/[file tail $file]
}
lappend filelist quedit
lappend filelist servers
lappend filelist VERSION

# If ~/.quirc doesn't exist, just create it and install the data files.

set quirc $env(HOME)/.quirc

if { ![file exists $quirc] } {
    cp $DATADIR/*.tcl $DATADIR/quedit $DATADIR/servers $DATADIR/VERSION $quirc
    cp $DATADIR/themes/*.tcl $quirc/themes
    cp $DATADIR/*.tcl $DATADIR/quedit $DATADIR/servers $DATADIR/VERSION $quirc/.originals
    cp $DATADIR/themes/*.tcl $quirc/.originals/themes
    exec chmod -f u+x $quirc/quedit
    log "Copied files from:"
    log "  $DATADIR"
    log "to:"
    log "  $quirc"
    log "and to:"
    log "  $quirc/.originals"
    return
}

set doinstall 0

if { $FOUNDI } { set doinstall 1 }

if { !$doinstall && ![file exists $quirc/variables.tcl] } {
    log "$quirc/variables.tcl is missing."
    log "Running installation routine."
    set doinstall 1
}

if { !$doinstall && ![file exists $quirc/VERSION] } {
    log "$quirc/VERSION is missing."
    log "Running installation routine."
    set doinstall 1
}

if { !$doinstall && [string compare [exec cat $quirc/VERSION] [version]] } {
    log "The contents of $quirc/VERSION do not match [version]."
    log "Running installation routine."
    set doinstall 1
}

if { !$doinstall } { destroy . }

# Back up old files.
file delete -force $quirc/backup
log "Removed old (if any) $quirc/backup."
log ""
cp $quirc/*.tcl $quirc/quedit $quirc/servers $quirc/VERSION $quirc/backup
cp $quirc/themes/*.tcl $quirc/backup/themes
log "Copied files from:"
log "  $quirc"
log "to:"
log "  $quirc/backup"
log ""

if { [tk_dialog .type "Installation Type" "Please choose the type of installation you would like.  The easy install will simply make a backup of your existing data files and install the new ones.  The advanced install will attempt to determine what needs upgraded and will ask you how you would like to deal with it.  It is more complex." "" 0 "Easy"] == 0 } {
    # Copy over new files
    cp $DATADIR/*.tcl $DATADIR/quedit $DATADIR/servers $DATADIR/VERSION $quirc
    cp $DATADIR/themes/*.tcl $quirc/themes
    cp $DATADIR/*.tcl $DATADIR/quedit $DATADIR/servers $DATADIR/VERSION $quirc/.originals
    cp $DATADIR/themes/*.tcl $quirc/.originals/themes
    exec chmod -f u+x $quirc/quedit
    log "Copied files from:"
    log "  $DATADIR"
    log "to:"
    log "  $quirc"
    log "and to:"
    log "  $quirc/.originals"
    log ""
    log "***** It would be in your best interest to read the following:"
    log ""
    log "The original files have been placed in:"
    log "  $quirc/backup"
    log ""
    log "The easy install does not check for extraneous files in the:"
    log "  $quirc"
    log "directory.  You will need to either check for yourself, pretend"
    log "that no such files exist, or run the advanced install; all of"
    log "which are viable options."
} else {
    # Options to see diff, and both files.

    # n is normal.  Not shown in list.
    # e is extraneous:
    #  Delete Extraneous
    #  ALL - Delete Extraneous
    # m is modified with no original:
    #  Overwrite
    #  ALL - Overwrite Modified Files With No Original Counterpart
    #  ALL - Overwrite
    # o is modified with original:
    #  Overwrite
    #  Merge
    #  Merge And Replace If No Conflicts Found
    #  ALL - Overwrite Modified Files That Have An Original Counterpart
    #  ALL - Overwrite
    #  ALL - Merge
    #  ALL - Merge And Replace If No Conflicts Found
    
    # Okay, so we want to merge.  They click on Merge.  Up pops a window with
    # the merged output.  The window allows editing and saving of the file, or
    # of killing of the editor.  If the safe option is chosen, it will not edit
    # files that merge okay, it will just replace them.

    # 	   -[M]erge The Files.  If the merge does not encounter any conflicts,
    #       the file will be named `basename $file`.merged.  If it does
    #       encounter conflicts, it will be called `basename $file`.fixme.
    # 	   -Merge All The [F]iles
    # 	   -Merge With [R]eplacement If Safe.  If the merge does not encounter
    #       any conflicts, the file in ~/.quirc will simply be replaced.
    #       Otherwise, the file will be left as `basename $file`.fixme in
    #       ~/.quirc just as the first merge option."
    # 	   -Merge And Replace All If [S]afe."
    #  
    foreach file $filelist {
	set type($file) "n"
	set diff($file) ""
    }
    foreach file [glob $env(HOME)/.quirc/*] {
	if { ![file isdirectory $file] } {
	    if { [lsearch -exact -- $filelist [file tail $file]]==-1 } {
		set type([file tail $file]) "e"
	    }
	}
    }
    # Could display a list of files with status indicators beside them
    # indicating what the deal is.  Then buttons could be provided that would
    # manipulate the state of each file.
    # Don't forget about original's dir and checking for validation of files
    # that needed to be installed.
    # Possible states:
    #  -Does not exist in new version (But did exist in older version?)
    #         Could check diff and delete if the same.

    #  -Do not show correctly installed files
    #  -Copy over new files automatically and do not show them

    #  -Force fixing until all are correctly installed (or allow exit)


    # For each file have a few fields... "Extraneous," "Modified," and the diff

    # New steps:
    #  -Mark extraneous files and have "Delete" and "Delete All Extraneous."
    #  -Make sure new directories exist
    #  -Copy over new files to ~/.quirc
    #  -Does orignals dir exist?
    #    Yes:
    #     -Copy over new files to originals.
    #     -Remove new files from originals.
    #     -If the files are not the same between the originals and quirc as
    #      well as the data and quirc, then mark it as different and store the
    #      diff.
    #    No:
    #     -Check for differences between data and quirc.  If different, 

    # Old steps:
    #  -Make backup
    #  -Offer to delete extraneous files
    #  -Copy over new files to ~/.quirc
    #  -Copy over new files to ~/.quirc/.originals (if originals dir exists)
    #  -Does the orignals dir exist?
    #   Yes:
    #    -See if the file in the quirc and original files differ.  If not, just
    #     copy over.
    #    -If so, check if different between data and quirc.  Display diff.
    #    -Display options like:
    # 	   -[L]eave It Alone
    # 	   -Leave Them [A]ll Alone
    # 	   -[O]verwrite It
    # 	   -Overwrite All Of [T]hem
    # 	   -[M]erge The Files.  If the merge does not encounter any conflicts,
    #       the file will be named `basename $file`.merged.  If it does
    #       encounter conflicts, it will be called `basename $file`.fixme.
    # 	   -Merge All The [F]iles
    # 	   -Merge With [R]eplacement If Safe.  If the merge does not encounter
    #       any conflicts, the file in ~/.quirc will simply be replaced.
    #       Otherwise, the file will be left as `basename $file`.fixme in
    #       ~/.quirc just as the first merge option."
    # 	   -Merge And Replace All If [S]afe."
    #   No:
    #    -See if the data files and the quirc files are different
    #    -If so, display diff and allow overwrite.
    #  -Replace the original files
    #  -chmod the quirc files
    log "Note that the advanced install does not perform checks to make sure"
    log "that all the necessary files have been upgraded.  You may need to"
    log "rerun the install using quirc -i, or edit some files manually if you"
    log "did not upgrade all the ncessary files."
}

log ""
log "If QuIRC seems to malfunction after the install, it may be because your"
log "default_scripts or default_globalscripts variables are not loading the"
log "correct scripts.  You should either delete those lines from"
log "~/.quirc/dynamic.tcl, remove the file entirely, or make the scripts match"
log "up appropriately with those listed in ~/.quirc/variables.tcl."

# Install script here using $DATADIR

# echo "This script is fairly new, and relatively untested, so please let me"
# echo "know about any problems you may have with it.  Thanks."
# echo ""
# echo "This script uses four different directories.  Their names and locations"
# echo "are as follows:"
# echo "     Data = $DATADIR"
# echo "     Backup = $HOME/.quirc/.backup"
# echo "     Originals = $HOME/.quirc/.originals"
# echo "     ~/.quirc = $HOME/.quirc"
# echo "Data contains the unmodified versions of the data files of the latest"
# echo "installed version of QuIRC."
# echo "Backup is created and used by the script as mentioned later."
# echo "Originals contains files that are used to attempt automatic merging of"
# echo "script files upon upgrading."
# echo "~/.quirc is the dir where the currently in use scripts for QuIRC are stored."
# echo ""
# echo "This script essentially copies the files from the data dir into ~/.quirc.  It"
# echo "does cleaning and merging as well, if requested.  The first thing it does is"
# echo "make a back-up copy of the files in ~/.quirc into the backup dir.  If the script"
# echo "fails for some reason, it is suggested that you recover your files, if"
# echo "necessary, from that directory, before running quirc.inst again."
# echo ""
# echo "Press Enter to continue..."
# read input

# echo "Making backup copy of ~/.quirc/* in ~/.quirc/.backup (Ignore directory errors)"
# echo ""
# rm -rf $HOME/.quirc/.backup
# mkdir $HOME/.quirc/.backup
# cp -f $HOME/.quirc/* $HOME/.quirc/.backup
# echo ""

# echo "Checking through each file in ~/.quirc.  If there is not a corresponding"
# echo "file in the data directory, offer to delete it."
# echo ""
# deleteall="n"
# for file in $HOME/.quirc/*; do
#     if [ ! -d $file ]; then
# 	if [ ! -f $DATADIR/`basename $file` ]; then
# 	    if test "$deleteall" = "y"; then
# 		echo "Deleting $file."
# 		rm $file
# 		continue
# 	    fi
# 	    echo "`basename $file` does not appear in the original data directory."
# 	    echo -n "Would you like to delete it ([Y]es/[N]o/Yes [T]o All/No To [A]ll)? "
# 	    read input
# 	    if test "$input" = "a" || test "$input" = "A"; then
# 		break
# 	    fi
# 	    if test "$input" = "t" || test "$input" = "T"; then
# 		echo "Deleting $file."
# 		rm $file
# 		deleteall="y"
# 	    fi
# 	    if test "$input" = "y" || test "$input" = "Y"; then
# 		echo "Deleting $file."
# 		rm $file
# 	    fi
# 	fi
#     fi
# done
# echo ""

# echo "Checking for files that are in the data dir but not in ~/.quirc, and"
# echo "copying them over."
# echo ""
# for file in $DATADIR/*; do
#     if [ ! -f $HOME/.quirc/`basename $file` ]; then
#         echo "Copying $file to ~/.quirc/`basename $file`"
# 	cp $file $HOME/.quirc/`basename $file`
#     fi
# done
# echo ""

# echo "Checking for new data files that don't appear in the original files"
# echo "directory, and copying them over."
# echo ""
# for file in $DATADIR/*; do
#     if [ ! -f $HOME/.quirc/.originals/`basename $file` ]; then
# 	echo "Copying $file to ~/.quirc/.originals"
# 	cp $file $HOME/.quirc/.originals
#     fi
# done
# echo ""

# if [ ! -d $HOME/.quirc/.originals ]; then
#     echo "~/.quirc/.originals doesn't exist.  Going through the data files"
#     echo "from the data dir, and seeing if they exist in ~/.quirc and are"
#     echo "different from each other. If so, offer to overwrite.  The unified"
#     echo "diff will be displayed for any files that are different."
#     echo ""
#     echo "Press Enter to continue..."
#     read input
#     overwriteall="n"
#     for file in $DATADIR/*; do
# 	if ! diff -c $HOME/.quirc/`basename $file` $file > /dev/null; then
# 	    if test "$overwriteall" = "y"; then
# 		echo "Overwriting ~/.quirc/`basename $file` with $file."
# 		cp -f $file $HOME/.quirc
# 		continue
# 	    fi
# 	    if ! diff -c $HOME/.quirc/`basename $file` $file; then
# 		echo "~/.quirc/`basename $file` is different from $file."
# 		echo -n "Overwrite it ([Y]es/[N]o/Yes [T]o All/No To [A]ll)? "
# 		read input
# 		if test "$input" = "a" || test "$input" = "A"; then
# 		    break
# 		fi
# 		if test "$input" = "t" || test "$input" = "T"; then
# 		    echo "Overwriting ~/.quirc/`basename $file` with $file."
# 		    cp -f $file $HOME/.quirc
# 		    overwriteall="y"
# 		fi
# 		if test "$input" = "y" || test "$input" = "Y"; then
# 		    echo "Overwriting ~/.quirc/`basename $file` with $file."
# 		    cp -f $file $HOME/.quirc
# 		fi
# 	    fi
# 	fi
#     done
#     echo ""
# else
#     overwriteall="n"
#     mergeall="n"
#     safemergeall="n"
#     echo "For each file in the data dir, check if the corresponding file in"
#     echo "the ~/.quirc dir is different from the corresponding file in the"
#     echo "originals dir.  If it isn't, overwrite it with the ne data file.  If"
#     echo "it is, check for differences between the old file in ~/.quirc and"
#     echo "the new data file.  Unified diff results will be displayed if so."
#     echo ""
#     # Go through the data files.
#     for file in $DATADIR/*; do
# 	# If the original file is the same as the ~/.quirc file, copy the data
#         # file to ~/.quirc.
# 	if diff -c $HOME/.quirc/`basename $file` $HOME/.quirc/.originals/`basename $file` > /dev/null; then
# 	    echo "Copying $file to $HOME/.quirc."
# 	    cp -f $file $HOME/.quirc
# 	    continue
# 	fi
# 	# If the files in ~/.quirc and the data dir are different, do the "all"
#         # options presently set.
# 	if ! diff -c $HOME/.quirc/`basename $file` $file > /dev/null; then
# 	    if test "$overwriteall" = "y"; then
# 		echo "Overwriting ~/.quirc/`basename $file` with $file."
# 		cp -f $file $HOME/.quirc
# 		continue
# 	    fi
# 	    if test "$mergeall" = "y"; then
# 		if diff3 -m $HOME/.quirc/`basename $file` $HOME/.quirc/.originals/`basename $file` $file > $HOME/.quirc/`basename $file`.fixme; then
# 		    mv -f $HOME/.quirc/`basename $file`.fixme $HOME/.quirc/`basename $file`.merged
# 		    echo "Merged to ~/.quirc/`basename $file`.merged"
# 		else
# 		    echo "Merged to ~/.quirc/`basename $file`.fixme"
# 		fi
# 		continue
# 	    fi
# 	    if test "$safemergeall" = "y"; then
# 		if diff3 -m $HOME/.quirc/`basename $file` $HOME/.quirc/.originals/`basename $file` $file > $HOME/.quirc/`basename $file`.fixme; then
# 		    mv -f $HOME/.quirc/`basename $file`.fixme $HOME/.quirc/`basename $file`
# 		    echo "Merged to ~/.quirc/`basename $file`"
# 		else
# 		    echo "Merged to ~/.quirc/`basename $file`.fixme"
# 		fi
# 		continue
# 	    fi
# 	fi
# 	# Do a visible diff of the files in ~/.quirc and the data dir.  If
#         # different, offer alternatives.
# 	if ! diff -c $file $HOME/.quirc/`basename $file`; then
# 	    echo ""
# 	    echo "`basename $file` in the originals dir is different from that in the data dir."
# 	    echo "What would you like to do with the existing file in ~/.quirc?"
# 	    echo "  [L]eave It Alone"
# 	    echo "  Leave Them [A]ll Alone"
# 	    echo "  [O]verwrite It"
# 	    echo "  Overwrite All Of [T]hem"
# 	    echo "  [M]erge The Files.  If the merge does not encounter any conflicts, the"
# 	    echo "    file will be named `basename $file`.merged.  If it does encounter conflicts,"
# 	    echo "    it will be called `basename $file`.fixme."
# 	    echo "  Merge All The [F]iles"
# 	    echo "  Merge With [R]eplacement If Safe.  If the merge does not encounter any"
# 	    echo "    conflicts, the file in ~/.quirc will simply be replaced.  Otherwise, the"
# 	    echo "    file will be left as `basename $file`.fixme in ~/.quirc just as the first"
# 	    echo "    merge option."
# 	    echo "  Merge And Replace All If [S]afe."
# 	    echo -n "> "
# 	    read input
# 	    if test "$input" = "a" || test "$input" = "A"; then
# 		break
# 	    fi
# 	    if test "$input" = "o" || test "$input" = "O"; then
# 		echo "Overwriting ~/.quirc/`basename $file` with $file."
# 		cp -f $file $HOME/.quirc
# 	    fi
# 	    if test "$input" = "t" || test "$input" = "T"; then
# 		overwriteall="y"
# 		echo -n "Overwriting ~/.quirc/`basename $file` with $file."
# 		cp -f $file $HOME/.quirc
# 	    fi
# 	    if test "$input" = "m" || test "$input" = "M"; then
# 		if diff3 -m $HOME/.quirc/`basename $file` $HOME/.quirc/.originals/`basename $file` $file > $HOME/.quirc/`basename $file`.fixme; then
# 		    mv -f $HOME/.quirc/`basename $file`.fixme $HOME/.quirc/`basename $file`.merged
# 		    echo "Merged to ~/.quirc/`basename $file`.merged"
# 		else
# 		    echo "Merged to ~/.quirc/`basename $file`.fixme"
# 		fi
# 	    fi
# 	    if test "$input" = "f" || test "$input" = "F"; then
# 		if diff3 -m $HOME/.quirc/`basename $file` $HOME/.quirc/.originals/`basename $file` $file > $HOME/.quirc/`basename $file`.fixme; then
# 		    mv -f $HOME/.quirc/`basename $file`.fixme $HOME/.quirc/`basename $file`.merged
# 		    echo -n "Merged to ~/.quirc/`basename $file`.merged"
# 		else
# 		    echo -n "Merged to ~/.quirc/`basename $file`.fixme"
# 		fi
# 		mergeall="y"
# 	    fi
# 	    if test "$input" = "r" || test "$input" = "R"; then
# 		if diff3 -m $HOME/.quirc/`basename $file` $HOME/.quirc/.originals/`basename $file` $file > $HOME/.quirc/`basename $file`.fixme; then
# 		    mv -f $HOME/.quirc/`basename $file`.fixme $HOME/.quirc/`basename $file`
# 		    echo "Merged to ~/.quirc/`basename $file`"
# 		else
# 		    echo "Merged to ~/.quirc/`basename $file`.fixme"
# 		fi
# 	    fi
# 	    if test "$input" = "s" || test "$input" = "S"; then
# 		if diff3 -m $HOME/.quirc/`basename $file` $HOME/.quirc/.originals/`basename $file` $file > $HOME/.quirc/`basename $file`.fixme; then
# 		    mv -f $HOME/.quirc/`basename $file`.fixme $HOME/.quirc/`basename $file`
# 		    echo -n "Merged to ~/.quirc/`basename $file`"
# 		else
# 		    echo -n "Merged to ~/.quirc/`basename $file`.fixme"
# 		fi
# 		safemergeall="y"
# 	    fi
# 	    echo ""
# 	else
# 	    echo "Copying $file to ~/.quirc"
# 	    cp $file $HOME/.quirc
# 	fi
#     done
#     echo ""
# fi

# echo "Replace the files in ~/.quirc/.originals with the data files for the"
# echo -n "version which you are installing ([Y]es/[N]o)? "
# read input
# if test "$input" = "y" || test "$input" = "Y"; then
#     if [ ! -d $HOME/.quirc/.originals ]; then
# 	mkdir $HOME/.quirc/.originals
#     fi
#     rm $HOME/.quirc/.originals/*
#     cp $DATADIR/* $HOME/.quirc/.originals
# fi
# echo ""

# #diff3 MINE OLDER YOURS
# #diff3 -m file.custom file.quirc.older file.quirc.newer
# #$? is 0 if succesful, 1 if not.

# echo "Running chmod on the necessary data files in ~/.quirc in order to"
# echo "ensure that they are useable."
# for file in $DATADIR/*; do
#   chmod -f u+r $HOME/.quirc/`basename $file`
# done
# chmod -f u+x $HOME/.quirc/quedit
# echo ""

# echo "If you got any files that were merged to .fixme files, you will need to"
# echo "edit those files, locate the conflicts, and resolve those manually."
# echo "For more information about the merging process, see your documentation"
# echo "for diff3."
# echo ""
# echo "To start QuIRC, enter quirc at the command line."

# #    echo "Checking for original files that aren't included in the new set of"
# #    echo "data files, and deleting them."
# #    echo ""
# #    for file in $HOME/.quirc/.originals/*; do
# #	if [ ! -f $DATADIR/`basename $file` ]; then
# #	    echo "Deleting $file."
# #	    rm $file
# #	fi
# #    done
# #    echo ""
